@@ -24,29 +24,29 @@
 module Network.Protocol.HTTP.DAV (
     DAVT(..)
   , evalDAVT
+  , withDAVContext
+  , runDAVContext
   , setCreds
   , setDepth
   , setResponseTimeout
   , setUserAgent
   , DAVContext(..)
-  , getProps
-  , getPropsAndContent
-  , putContentAndProps
-  , putContent
-  , deleteContent
-  , moveContent
-  , makeCollection
-  , caldavReport
   , caldavReportM
   , delContentM
   , getPropsM
   , getContentM
+  , withContentM
   , mkCol
   , moveContentM
   , putPropsM
   , putContentM
+  , putContentM'
   , withLockIfPossible
   , withLockIfPossibleForDelete
+  , inDAVLocation
+  , getDAVLocation
+  , mkDAVContext
+  , closeDAVContext
   , module Network.Protocol.HTTP.DAV.TH
 ) where
 
@@ -54,25 +54,28 @@ import Network.Protocol.HTTP.DAV.TH
 
 import Control.Applicative (liftA2, Applicative)
 import Control.Error (EitherT(..))
-import Control.Exception.Lifted (catchJust, finally, bracketOnError)
-import Control.Lens ((^.), (.=), (%=))
-import Control.Monad (liftM, liftM2, when, MonadPlus)
+import Control.Lens ((^.), (.=), (%=), (.~))
+import Control.Monad (when, MonadPlus)
 import Control.Monad.Base (MonadBase(..))
+import Control.Monad.Catch (bracket, bracketOnError, catchJust, finally, throwM, mask, uninterruptibleMask, MonadCatch, MonadMask, MonadThrow)
+import qualified Control.Monad.Catch as MonadCatch
 import Control.Monad.Error (MonadError)
 import Control.Monad.Fix (MonadFix)
-import Control.Monad.Trans (lift, MonadTrans)
+import Control.Monad.Trans.Class (lift, MonadTrans)
+import Control.Monad.Trans.Either (left)
 import Control.Monad.IO.Class (liftIO, MonadIO)
-import Control.Monad.State (evalStateT, get, MonadState, StateT)
-import Control.Monad.Trans.Control (MonadBaseControl(..))
+import Control.Monad.State (evalStateT, runStateT, get, MonadState, StateT)
 
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as BC8
 import qualified Data.ByteString.Lazy as BL
+import qualified Data.ByteString.UTF8 as UTF8B
+import Data.Default (Default, def)
 import qualified Data.Map as Map
 
 import Data.Maybe (catMaybes, fromMaybe)
 
-import Network.HTTP.Client (RequestBody(..), httpLbs, parseUrl, applyBasicAuth, Request(..), Response(..), newManager, closeManager, HttpException(..))
+import Network.HTTP.Client (RequestBody(..), httpLbs, parseUrl, applyBasicAuth, Request(..), Response(..), newManager, closeManager, HttpException(..), BodyReader, withResponse, path)
 import Network.HTTP.Client.TLS (tlsManagerSettings)
 import Network.HTTP.Types (hContentType, Method, Status, RequestHeaders, unauthorized401, conflict409)
 
@@ -82,31 +85,57 @@ import Text.Hamlet.XML (xml)
 
 import Data.CaseInsensitive (mk)
 
+instance Default DAVContext where
+    def = DAVContext [] def B.empty B.empty [] Nothing def Nothing "hDav-using application"
+
 newtype DAVT m a = DAVT { runDAVT :: EitherT String (StateT DAVContext m) a }
     deriving (Applicative, Functor, Monad, MonadBase b, MonadError String, MonadFix, MonadIO, MonadPlus, MonadState DAVContext)
 
-instance MonadBaseControl b m => MonadBaseControl b (DAVT m) where
-   newtype StM (DAVT m) a = StDAVT { unStDAVT :: StM (EitherT String (StateT DAVContext m)) a }
-   liftBaseWith f = DAVT . liftBaseWith $ \r -> f $ liftM StDAVT . r . runDAVT
-   restoreM       = DAVT . restoreM . unStDAVT
+-- this orphan instance is probably a bad idea
+instance MonadMask m => MonadMask (EitherT e m) where
+    mask a = EitherT $ mask $ \u -> runEitherT (a $ q u)
+        where q u = EitherT . u . runEitherT
+    uninterruptibleMask a = EitherT $ uninterruptibleMask $ \u -> runEitherT (a $ q u)
+        where q u = EitherT . u . runEitherT
+
+instance MonadCatch m => MonadCatch (DAVT m) where
+    catch (DAVT m) f = DAVT $ MonadCatch.catch m (runDAVT . f)
+
+instance MonadMask m => MonadMask (DAVT m) where
+    mask a = DAVT $ mask $ \u -> runDAVT (a $ q u)
+        where q u = DAVT . u . runDAVT
+    uninterruptibleMask a = DAVT $ uninterruptibleMask $ \u -> runDAVT (a $ q u)
+        where q u = DAVT . u . runDAVT
+
+instance MonadThrow m => MonadThrow (DAVT m) where
+    throwM = lift . throwM
 
 instance MonadTrans DAVT where
     lift = DAVT . lift . lift
 
-evalDAVT :: MonadIO m => String -> DAVT m a -> m (Either String a)
+type DAVURL = String
+
+evalDAVT :: MonadIO m => DAVURL -> DAVT m a -> m (Either String a)
 evalDAVT u f = do
-    mgr <- liftIO $ newManager tlsManagerSettings
-    req <- liftIO $ parseUrl u
-    r <- (evalStateT . runEitherT . runDAVT) f $ DAVContext [] req B.empty B.empty [] Nothing mgr Nothing "hDav-using application"
-    liftIO $ closeManager mgr
+    ctx <- mkDAVContext u
+    r <- (evalStateT . runEitherT . runDAVT) f ctx
+    closeDAVContext ctx
     return r
 
-choke :: IO (Either String a) -> IO a
-choke f = do
-   x <- f
-   case x of
-       Left e -> error e
-       Right r -> return r
+mkDAVContext :: MonadIO m => DAVURL -> m DAVContext
+mkDAVContext u = liftIO $ do
+    mgr <- liftIO $ newManager tlsManagerSettings
+    req <- liftIO $ parseUrl u
+    return $ def { _baseRequest = req, _httpManager = Just mgr }
+
+closeDAVContext :: MonadIO m => DAVContext -> m ()
+closeDAVContext ctx = liftIO $ maybe (return ()) closeManager (ctx ^. httpManager)
+
+withDAVContext :: (MonadIO m, MonadMask m) => DAVURL -> (DAVContext -> m a) -> m a
+withDAVContext u = bracket (mkDAVContext u) closeDAVContext
+
+runDAVContext :: MonadIO m => DAVContext -> DAVT m a -> m (Either String a, DAVContext)
+runDAVContext ctx f = (runStateT . runEitherT . runDAVT) f ctx
 
 setCreds :: MonadIO m => B.ByteString -> B.ByteString -> DAVT m ()
 setCreds u p = basicusername .= u >> basicpassword .= p
@@ -120,8 +149,8 @@ setUserAgent ua = userAgent .= ua
 setResponseTimeout :: MonadIO m => Maybe Int -> DAVT m ()
 setResponseTimeout rt = baseRequest %= \x -> x { responseTimeout = rt }
 
-davRequest :: MonadIO m => Method -> RequestHeaders -> RequestBody -> DAVT m (Response BL.ByteString)
-davRequest meth addlhdrs rbody = do
+mkDavRequest :: MonadIO m => Method -> RequestHeaders -> RequestBody -> DAVT m Request
+mkDavRequest meth addlhdrs rbody = do
     ctx <- get
     let hdrs = catMaybes
                [ Just (mk "User-Agent", ctx ^. userAgent)
@@ -131,7 +160,14 @@ davRequest meth addlhdrs rbody = do
         authreq = if B.null (ctx ^. basicusername) && B.null (ctx ^. basicpassword)
             then req
             else applyBasicAuth (ctx ^. basicusername) (ctx ^. basicpassword) req
-    liftIO (httpLbs authreq (ctx ^. httpManager))
+    return authreq
+
+davRequest :: MonadIO m => Method -> RequestHeaders -> RequestBody -> DAVT m (Response BL.ByteString)
+davRequest meth addlhdrs rbody = go =<< mkDavRequest meth addlhdrs rbody
+  where
+    go req = do
+      ctx <- get
+      maybe (DAVT $ left "Can't perform request without manager") (liftIO . httpLbs req) (ctx ^. httpManager)
 
 matchStatusCodeException :: Status -> HttpException -> Maybe ()
 matchStatusCodeException want (StatusCodeException s _ _)
@@ -182,18 +218,33 @@ getPropsM = do
     propresp <- davRequest "PROPFIND" ahs (xmlBody propname)
     return $ (XML.parseLBS_ XML.def . responseBody) propresp
 
+-- | Note that the entire request body is buffered in memory.
+-- To stream large files use withContentM instead.
 getContentM :: MonadIO m => DAVT m (Maybe B.ByteString, BL.ByteString)
 getContentM = do
     resp <- davRequest "GET" [] emptyBody
     let ct = lookup hContentType (responseHeaders resp)
     return (ct, responseBody resp)
 
+withContentM :: MonadIO m => (Response BodyReader -> IO a) -> DAVT m a
+withContentM handleresponse = do
+    req <- mkDavRequest "GET" [] emptyBody
+    ctx <- get
+    maybe (DAVT $ left "Can't handle response without manager") (\mgr -> liftIO $ withResponse req mgr handleresponse) (ctx ^. httpManager)
+
+-- | Note that the entire request body is buffered in memory; not suitable
+-- for large files.
 putContentM :: MonadIO m => (Maybe B.ByteString, BL.ByteString) -> DAVT m ()
-putContentM (ct, body) = do
+putContentM (ct, body) = putContentM' (ct, RequestBodyLBS body)
+
+-- | To send a large file, pass eg a RequestBodyStream containing the
+-- file's content.
+putContentM' :: MonadIO m => (Maybe B.ByteString, RequestBody) -> DAVT m ()
+putContentM' (ct, requestbody) = do
     d <- get
     let ahs' = maybe [] (return . (,) (mk "If") . parenthesize) (d ^. lockToken)
     let ahs = ahs' ++ maybe [] (return . (,) hContentType) ct
-    _ <- davRequest "PUT" ahs (RequestBodyLBS body)
+    _ <- davRequest "PUT" ahs requestbody
     return ()
 
 delContentM :: MonadIO m => DAVT m ()
@@ -212,7 +263,7 @@ mkCol' = do
     _ <- davRequest "MKCOL" [] emptyBody
     return ()
 
-mkCol :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => DAVT m Bool
+mkCol :: (MonadIO m, MonadBase IO m, MonadCatch m) => DAVT m Bool
 mkCol = catchJust
         (matchStatusCodeException conflict409)
         (mkCol' >> return True)
@@ -260,14 +311,14 @@ caldavReportM = do
 getOptionsOnce :: MonadIO m => DAVT m ()
 getOptionsOnce = getOptions -- this should only happen once
 
-withLockIfPossible :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => Bool -> DAVT m a -> DAVT m a
+withLockIfPossible :: (MonadIO m, MonadBase IO m, MonadCatch m, MonadMask m) => Bool -> DAVT m a -> DAVT m a
 withLockIfPossible nocreate f = do
     getOptionsOnce
     o <- get
     when (supportsLocking o) (lockResource nocreate)
     f `finally` when (supportsLocking o) unlockResource
 
-withLockIfPossibleForDelete :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => Bool -> DAVT m a -> DAVT m a
+withLockIfPossibleForDelete :: (MonadIO m, MonadBase IO m, MonadCatch m, MonadMask m) => Bool -> DAVT m a -> DAVT m a
 withLockIfPossibleForDelete nocreate f = do
     getOptionsOnce
     o <- get
@@ -276,63 +327,6 @@ withLockIfPossibleForDelete nocreate f = do
     let unlock = when (supportsLocking o) unlockResource
     bracketOnError lock (const unlock) (const f)
 
-{-# DEPRECATED getProps "This function will be removed in favor of getPropsM" #-}
-getProps :: String -> B.ByteString -> B.ByteString -> Maybe Depth -> IO XML.Document
-getProps url username password md = choke $ evalDAVT url $ do
-    setCreds username password
-    setDepth md
-    getPropsM
-
-{-# DEPRECATED getPropsAndContent "This function will be removed in favor of getPropsM and getContentM" #-}
-getPropsAndContent :: String -> B.ByteString -> B.ByteString -> IO (XML.Document, (Maybe B.ByteString, BL.ByteString))
-getPropsAndContent url username password = choke $ evalDAVT url $ do
-    setCreds username password
-    setDepth (Just Depth0)
-    withLockIfPossible True $ liftM2 (,) getPropsM getContentM
-
-{-# DEPRECATED putContent "This function will be removed in favor of putContentM" #-}
-putContent :: String -> B.ByteString -> B.ByteString -> (Maybe B.ByteString, BL.ByteString) -> IO ()
-putContent url username password b = choke $ evalDAVT url $ do
-    setCreds username password
-    withLockIfPossible False $ putContentM b
-
-{-# DEPRECATED putContentAndProps "This function will be removed in favor of putContentM and putPropsM" #-}
-putContentAndProps :: String -> B.ByteString -> B.ByteString -> (XML.Document, (Maybe B.ByteString, BL.ByteString)) -> IO ()
-putContentAndProps url username password (p, b) = choke $ evalDAVT url $ do
-    setCreds username password
-    withLockIfPossible False $ do putContentM b
-                                  putPropsM p
-
-{-# DEPRECATED deleteContent "This function will be removed in favor of delContentM" #-}
-deleteContent :: String -> B.ByteString -> B.ByteString -> IO ()
-deleteContent url username password = choke $ evalDAVT url $ do
-    setCreds username password
-    withLockIfPossibleForDelete False delContentM
-
-{-# DEPRECATED moveContent "This function will be removed in favor of moveContentM" #-}
-moveContent :: String -> B.ByteString -> B.ByteString -> B.ByteString -> IO ()
-moveContent url newurl username password = choke $ evalDAVT url $ do
-    setCreds username password
-    moveContentM newurl
-
-{-# DEPRECATED caldavReport "This function will be removed in favor of caldavReportM" #-}
-caldavReport :: String -> B.ByteString -> B.ByteString -> IO XML.Document
-caldavReport url username password = choke $ evalDAVT url $ do
-   setCreds username password
-   setDepth (Just Depth1)
-   caldavReportM
-
--- | Creates a WebDAV collection, which is similar to a directory.
---
--- Returns False if the collection could not be made due to an intermediate
--- collection not existing. (Ie, collection /a/b/c/d cannot be made until
--- collection /a/b/c exists.)
-{-# DEPRECATED makeCollection "This function will be removed in favor of mkCol" #-}
-makeCollection :: String -> B.ByteString -> B.ByteString -> IO Bool
-makeCollection url username password = choke $ evalDAVT url $ do
-    setCreds username password
-    mkCol
-
 propname :: XML.Document
 propname = XML.Document (XML.Prologue [] Nothing []) root []
     where
@@ -361,3 +355,39 @@ calendarquery = XML.Document (XML.Prologue [] Nothing []) root []
 <C:filter>
   <C:comp-filter name="VCALENDAR">
 |]
+
+-- | Normally, DAVT actions act on the url that is provided to eg, evalDAVT.
+-- Sometimes, it's useful to adjust the url that is acted on, while
+-- remaining in the same DAV session.
+--
+-- inLocation temporarily adjusts the url's path, while performing a
+-- DAVT action.
+--
+-- For example:
+--
+-- > import System.FilePath.Posix -- posix for url path manipulation
+-- >
+-- > mkColRecursive d = do
+-- >   let parent = takeDirectory d
+-- >   when (parent /= d) $
+-- >     mkColRecursive parent
+-- >   inDAVLocation (</> d) mkCol
+--
+-- Note that operations that modify the DAVContext
+-- (such as setCreds and setCreds) can be run inside davLocation,
+-- but will not have any effect on the calling DAVContext.
+inDAVLocation :: MonadIO m => (String -> String) -> DAVT m a -> DAVT m a
+inDAVLocation f a = do
+    ctx <- get
+    let r = ctx ^. baseRequest
+        r' = r { path = adjustpath r }
+        ctx' = baseRequest .~ r' $ ctx
+    lift $ either error return =<< (evalStateT . runEitherT . runDAVT) a ctx'
+  where
+    adjustpath = UTF8B.fromString . f . UTF8B.toString . path
+
+-- | Gets the path of the url that DAVT actions will act on.
+getDAVLocation :: Monad m => DAVT m String
+getDAVLocation = do
+    ctx <- get
+    return (UTF8B.toString $ path $ ctx ^. baseRequest)
